#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc

import numpy


class SummaryExperiment(epyc.Experiment):
    '''An experiment combinator that takes an underlying "base" experiment and 
    returns summary statistics for some of its results. This only really makes
    sense for experiments that return lists of results, such as those conducted
    using RepeatedExperiment.

    When run, a summary experiment summarises the results of the
    results, creating a new set of results that include the mean and
    variance for each result that the underyling experiments
    generated. (You can also select which results to summarise.) The
    raw results are discarded. The new results have the names of the
    raw results, suffixed by "_mean" and "_variance".

    The summary calculations only include those experimental runs that succeeded,
    that is that have their status set to True. Failed runs are ignored.'''

    # Additional metadata
    UNDERLYING_RESULTS = 'repetitions'                         # repetitions summarised
    UNDERLYING_SUCCESSFUL_RESULTS = 'successful_repetitions'   # repetitions included
                                                               # in the summary

    # Prefix and suffix tags attached to summarised result and metadata values
    MEAN_SUFFIX = '_mean'
    MEDIAN_SUFFIX = '_median'
    VARIANCE_SUFFIX = '_variance'
    
    
    def __init__( self, ex, summarised_results = None ):
        '''Create a summarised version of the given experiment. The given
        fields in the experimental results will be summarised, defaulting to all.

        ex: the underlying experiment
        summarised_results: list of result values to summarise (defaults to all)'''
        super(epyc.SummaryExperiment, self).__init__()
        self._experiment = ex
        self._summarised_results = summarised_results

    def experiment( self ):
        '''Return the underlying experiment.

        returns: the underlying experiment'''
        return self._experiment

    def _mean( self, k ):
        '''Return the tag associated with the mean of k.'''
        return k + self.MEAN_SUFFIX

    def _median( self, k ):
        '''Return the tag associated with the median of k.'''
        return k + self.MEDIAN_SUFFIX

    def _variance( self, k ):
        '''Return the tag associated with the variance of k.'''
        return k + self.VARIANCE_SUFFIX
    
    def summarise( self, results ):
        '''Generate a summary result dict from a list of result dicts
        returned by do() on the repetitions of the underlying experiment.
        By default we generate order, mean and variance for each value recorded.

        We drop from the calculations any experiments whose completion status
        was False, indicating an error.

        Override this method to create different summary statistics.

        results: an array of result dicts
        returns: a dict of summary statistics'''
        if len(results) == 0:
            return dict()
        else:
            summary = dict()

            # work out the fields to summarise
            ks = self._summarised_results
            if ks is None:
                ks = results[0][epyc.Experiment.RESULTS].keys()

            # add the summary statistics
            for k in ks:
                vs = [ res[epyc.Experiment.RESULTS][k] for res in results ]
                summary[self._mean(k)]     = numpy.mean(vs)
                summary[self._median(k)]   = numpy.median(vs)
                summary[self._variance(k)] = numpy.var(vs)

            return summary   

    def do( self, params ):
        '''Perform the underlying experiment and summarise its results.
        Our results are the summary statistics extracted from the results of
        the instances of the underlying experiment that we performed.

        We drop from the calculations any experiments whose completion status
        was False, indicating an error.

        params: the parameters to the underlying experiment
        returns: the summary statistics of the underlying results'''

        # set the parameters of the underlying experiment
        self.experiment().set(params)

        # perform the underlying experiment
        results = self.experiment().run()

        # extract only the successful runs
        sresults = [ res for res in results if res[epyc.Experiment.METADATA][epyc.Experiment.STATUS] ]

        # add extra values to out metadata record
        self._metadata[self.UNDERLYING_RESULTS]            = len(results)
        self._metadata[self.UNDERLYING_SUCCESSFUL_RESULTS] = len(sresults)
        
        # construct summary results
        return self.summarise(sresults)




# Base class for experiments
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from datetime import datetime, timedelta


class Experiment(object):
    '''Base class for experiments conducted in a lab.

    From the developer's perspective, an experiment has five methods
    that can be overridden: configure(), deconfigure(), setUp(), do(),
    and tearDown(). configure() sets up the experiment with its parameters.
    If parameters have been set previously, deconfigure() is called first.
    do() performs the experiment, and is bracketed by calls to setUp()
    and tearDown(). The idea is that configure() is called whenever the
    parameters to the experiment are changed, while the other methods
    can be called one or more times between resets.

    From an experimenter's (or a lab's) perspective, an experiment
    has public methods set() and run(). The former sets the parameters
    for the experiment; the latter runs the experiment. producing a set
    of results that include direct experimental results and metadata on
    how the experiment ran. A single run may produce a list of result
    dicts if desired, each filled-in with the correct metadata.

    Results can be accessed by the results() method. The experiment also
    exposes an indexing interface to direct experimental results.'''

    # Top-level structure for results
    METADATA = 'metadata'                 # metadata values, mainly timing
    PARAMETERS = 'parameters'             # point in the parameter space the experiment ran on
    RESULTS = 'results'                   # results generated at that point

    # Common metadata elements reported
    START_TIME = 'start_time'             # datetime experiment started
    END_TIME = 'end_time'                 # datetime experiment ended
    ELAPSED_TIME = 'elapsed_time'         # time experiment took overall in ms
    SETUP_TIME = 'setup_time'             # time spent on setup in ms
    EXPERIMENT_TIME = 'experiment_time'   # time spent on experiment itself in ms
    TEARDOWN_TIME = 'teardown_time'       # time spent on teardown in ms
    STATUS = 'status'                     # True if experiment completed successfully
    EXCEPTION = 'exception'               # exception thrown if experiment failed

    
    def __init__( self ):
        '''Create a new experiment.'''
        self._metadata = dict()
        self._parameters = None
        self._results = None

    def setUp( self, params ):
        '''Set up the experiment. Default does nothing.

        params: the parameters of the experiment'''
        pass

    def tearDown( self ):
        '''Tear down the experiment. Default does nothing.'''
        pass

    def configure( self, params ):
        '''Configure the experiment for the given parameters.
        Default does nothing.

        params: the parameters'''
        pass

    def deconfigure( self ):
        '''De-configure the experiment prior to setting new parameters.
        Default does nothing.'''
        pass

    def set( self, params ):
        '''Set the parameters for the experiment.

        params: the parameters
        returns: the experiment'''
        if self._parameters is not None:
            self.deconfigure()
        self._parameters = params
        self.configure(params)
        return self

    def do( self, params ):
        '''Do the body of the experiment. This should be overridden
        by sub-classes.

        params: a dict of parameters for the experiment
        returns: a dict of results'''
        raise NotYetImplementedError('do()')

    def report( self, params, meta, res ):
        '''Return a dict of results. The default returns a dict with
        results keyed by self.RESULTS, the data point in the parameter space
        keyed by self.PARAMETERS, and timing and other metadata keyed
        by self.METADATA. Overriding this method can be used to record extra
        values, but be sure to call the base method as well.
 
        params: the parameters we ran under
        meta: the metadata for this run
        res: the direct experimental results from do()
        returns: a dict of extended results'''
        rc = dict()
        rc[self.PARAMETERS] = params.copy()
        rc[self.METADATA] = meta.copy()
        rc[self.RESULTS] = res
        return rc

    def run( self ):
        '''Run the experiment, using the parameters set using set().
        A "run" consists of calling setUp(), do(), and tearDown(),
        followed by collecting and storing (and returning) the
        experiment's results.

        returns: dict of reported results'''

        # perform the experiment protocol
        params = self._parameters
        self._metadata = dict()
        self._results = None
        res = None
        doneSetupTime = doneExperimentTime = doneTeardownTime = None
        try:
            # do the phases in order, recording the wallclock times at each phase
            startTime = datetime.now()
            self.setUp(params)
            doneSetupTime = datetime.now()
            res = self.do(params)
            doneExperimentTime = datetime.now() 
            self.tearDown()
            doneTeardownTime = datetime.now() 

            # record the various timings
            self._metadata[self.START_TIME] = startTime
            self._metadata[self.END_TIME] = doneTeardownTime
            self._metadata[self.ELAPSED_TIME] = (doneTeardownTime - startTime).total_seconds() / 1000.0
            self._metadata[self.SETUP_TIME] = (doneSetupTime - startTime).total_seconds() / 1000.0
            self._metadata[self.EXPERIMENT_TIME] = (doneExperimentTime - doneSetupTime).total_seconds() / 1000.0
            self._metadata[self.TEARDOWN_TIME] = (doneTeardownTime - doneExperimentTime).total_seconds() / 1000.0

            # set the success flag
            self._metadata[self.STATUS] = True
        except Exception as e:
            print "Caught exception in experiment: {e}".format(e = e)
            
            # decide on the cleanup actions that need doing
            if (doneSetupTime is not None) and (doneExperimentTime is None):
                # we did the setup and then failed in the experiment, so
                # we need to do the teardown
                try:
                    self.tearDown()
                except:
                    pass
                
            # set the failure flag and record the exception
            # (there will be no timing information recorded)
            self._metadata[self.STATUS] = False
            self._metadata[self.EXCEPTION] = e

        # report the results
        self._results = res
        return self.report(self.parameters(),
                           self.metadata(),
                           res)
    
    def __getitem__( self, k ):
        '''Return the given element of the experimental results. This only
        gives access to direct experimental results, not to parameters
        or metadata.

        k: the result key
        returns: the value'''
        if self._results is None:
            raise Exception("No results set")
        else:
            return (self.results())[k]
    
    def success( self ):
        '''Test whether the experiment has been run successfully. This will
        be False if the experiment hasn't been run, or if it's been run and
        failed (in which case the exception will be stored in the metadata).

        returns: True if the experiment has been run successfully'''
        if self.STATUS in self.metadata().keys():
            return (self.metadata())[self.STATUS]
        else:
            return False

    def results( self ):
        '''Return a complete results dict. Only really makes sense for
        recently-executed experimental runs.

        return: the results dict'''
        return self.report(self.parameters(),
                           self.metadata(),
                           self.experimentalResults())

    def experimentalResults( self ):
        '''Return the experimental results from our last run. This will
        be None if we haven't been run, or if we ran and failed.

        return: the experimental results or None'''
        return self._results
    
    def parameters( self ):
        '''Return the current experimental parameters, which will
        be None if none have been given by a call to set()

        returns: the parameters,'''
        return self._parameters

    def metadata( self ):
        '''Return the metadata we collected at out last execution, which
        will be None if we've not been executed and an empty dict if
        we're mid-run (i.e., if this method is called from do() for
        whatever reason).

        returns: the metadata'''
        return self._metadata
    
    
        
    



# Tests of summarising experiments combinator
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from epyc import *

import unittest
import numpy
import numpy.random


class SampleExperiment1(Experiment):
    
    def do( self, params ):
        return dict(result = params['x'],
                    dummy = 1)

class SampleExperiment2(Experiment):

    def __init__( self ):
        super(SampleExperiment2, self).__init__()
        self._allvalues = []
    
    def do( self, params ):
        v = numpy.random.random()
        self._allvalues.append(v)
        return dict(result = v)

    def values( self ):
        return self._allvalues

class SampleExperiment3(SampleExperiment2):

    def __init__( self ):
        super(SampleExperiment3, self).__init__()
        self._ran = 0
    
    def do( self, params ):
        v = numpy.random.random()
        if v < 0.5:
            # experiment succeeds
            self._allvalues.append(v)
            self._ran = self._ran + 1
            return dict(result = v)
        else:
            # experiment fails
            raise Exception("Failing")

    def ran( self ):
        return self._ran

    
class SummaryExperimentTests(unittest.TestCase):

    def setUp( self ):
        self._lab = Lab()
        
    def testRepetitionsOnePoint( self ):
        '''Test we can repeat an experiment at a single point'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment1()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        res = (self._lab.results())[0]

        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_RESULTS], N)
        
    def testRepetitionsMultiplePoint( self ):
        '''Test we can repeat an experiment across a parameter space'''
        N = 10
        
        self._lab['x'] = [ 5, 10, 15 ]
        
        e = SampleExperiment1()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        res = self._lab.results()
        self.assertEqual(len(res), len(self._lab['x']))

        for i in range(len(self._lab['x'])):
            self.assertTrue(res[i][Experiment.METADATA][Experiment.STATUS])
            self.assertIn(res[i][Experiment.PARAMETERS]['x'], self._lab['x'])
            self.assertEqual(res[i][Experiment.METADATA][SummaryExperiment.UNDERLYING_RESULTS], N)
            self.assertEqual(res[i][Experiment.RESULTS]['result_mean'],
                             res[i][Experiment.PARAMETERS]['x'])
            self.assertEqual(res[i][Experiment.RESULTS]['result_variance'], 0)

    def testSummary( self ):
        '''Test that summary statistics are created properly'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment2()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        res = (self._lab.results())[0]

        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.RESULTS]['result_mean'],
                         numpy.mean(e.values()))
        self.assertEqual(res[Experiment.RESULTS]['result_variance'],
                         numpy.var(e.values()))
       
    def testSelect( self ):
        '''Test we can select summary results'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment1()
        es = SummaryExperiment(RepeatedExperiment(e, N),
                               summarised_results = [ 'dummy' ])

        self._lab.runExperiment(es)
        res = (self._lab.results())[0]
        
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(len(res[Experiment.RESULTS].keys()), 3)
        self.assertIn('dummy_mean', res[Experiment.RESULTS].keys())
        self.assertEqual(res[Experiment.RESULTS]['dummy_mean'], 1)
        self.assertIn('dummy_median', res[Experiment.RESULTS].keys())
        self.assertEqual(res[Experiment.RESULTS]['dummy_median'], 1)
        self.assertIn('dummy_variance', res[Experiment.RESULTS].keys())
        self.assertEqual(res[Experiment.RESULTS]['dummy_variance'], 0)

    def testIgnoring( self ):
        '''Test that we ignore failed experiments'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment3()
        es = SummaryExperiment(RepeatedExperiment(e, N))

        self._lab.runExperiment(es)
        res = (self._lab.results())[0]
        
        self.assertTrue(res[Experiment.METADATA][Experiment.STATUS])
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_RESULTS], N)
        self.assertEqual(res[Experiment.METADATA][SummaryExperiment.UNDERLYING_SUCCESSFUL_RESULTS], e.ran())
        self.assertEqual(res[Experiment.RESULTS]['result_mean'],
                         numpy.mean(e.values()))
        self.assertEqual(res[Experiment.RESULTS]['result_variance'],
                         numpy.var(e.values()))
         



# Simulation "lab" experiment management, sequential version
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc

import collections


class Lab(object):
    '''A laboratory for computational experiments. The lab conducts an
    experiment at different points in a multi-dimensional parameter space.
    The default performs all the experiments locally; sub-classes exist
    to perform remote parallel experiments.

    A Lab stores its result in a notebook, an instance of LabNotebook. By
    default the ordinary Lab classes uses an in-memory notebook, essentially
    just a dict; sub-clasess use persistent notebooks to manage larger
    sets of experiments.'''

    def __init__( self, notebook = None ):
        '''Create an empty lab.
 
        notebook: the notebook used to store results (defaults to an empty LabNotebook)'''
        if notebook is None:
            self._notebook = epyc.LabNotebook()
        else:
            self._notebook = notebook
        self._parameters = dict()

    def addParameter( self, k, r ):
        '''Add a parameter to the experiment's parameter space. k is the
        parameter name, and r is its range.

        k: parameter name
        r: parameter range'''

        # if range is a single value, make it a list
        if not isinstance(r, collections.Iterable):
            r = [ r ]
        self._parameters[k] = r

    def parameters( self ):
        '''Return a list of parameter names.

        returns: a list of names'''
        return self._parameters.keys()

    def __len__( self ):
        '''The length of an experiment is the total number of data points
        that will be explored. 

        returns: the length of the experiment'''
        n = 1
        for p in self.parameters():
            n = n * len(self._parameters[p])
        return n
        
    def __getitem__( self, k ):
        '''Access a parameter range using array notation.

        k: parameter name
        returns: the parameter range'''
        return self._parameters[k]

    def __setitem__( self, k, r ):
        '''Add a parameter using array notation.

        k: the parameter name
        r: the parameter range'''
        return self.addParameter(k, r)

    def _crossProduct( self, ls ):
        '''Internal method to generate the cross product of all parameter
        values, creating the parameter space for the experiment.

        ls: an array of parameter names
        returns: list of dicts'''
        p = ls[0]
        ds = []
        if len(ls) == 1:
            # last parameter, convert range to a dict
            for i in self._parameters[p]:
                dp = dict()
                dp[p] = i
                ds.append(dp)
        else:
            # for other parameters, create a dict combining each value in
            # the range to a copy of the dict of other parameters
            ps = self._crossProduct(ls[1:])
            for i in self._parameters[p]:
                for d in ps:
                    dp = d.copy()
                    dp[p] = i
                    ds.append(dp)

        # return the complete parameter space
        return ds
           
    def parameterSpace( self ):
        '''Return the parameter space of the experiment as a list of dicts,
        with each dict mapping each parameter name to a value.

        returns: the parameter space as a list of dicts'''
        ps = self.parameters()
        if len(ps) == 0:
            return []
        else:
            return self._crossProduct(ps)
    
    def runExperiment( self, e ):
        '''Run an experiment over all the points in the parameter space.

        e: the experiment'''

        # create the parameter space
        ps = self.parameterSpace()

        # run the experiment at each point
        nb = self.notebook()
        for p in ps:
            # print "Running {p}".format(p = p)
            res = e.set(p).run()
            nb.addResult(res)

        # commit the results
        nb.commit()

    def notebook( self ):
        '''Return the notebook being used by this lab.

        returns: the notebook'''
        return self._notebook
    
    def results( self ):
        '''Retrieve the list of result dicts (each of which contains the
        point at which the experiment was evaluated to get this
        result).

        returns: a list of experimental results'''
        return self.notebook().results()
    
    def ready( self ):
        '''Test whether all the results are ready, that is none are
        pending.

        returns: True if the results are in'''
        return (len(self.notebook().pendingResults()) == 0)





# Simulation "lab notebook" for collecting results, JSON version
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc

import os
import json
from datetime import datetime
import dateutil.parser


class MetadataEncoder(json.JSONEncoder):
    '''Add support for encoding Python datetime objects within
    JSON objects, using the standard ISO string representation.'''

    def default( self, o ):
        '''If o is a datetime object, convert it to an ISO string.

        o: the field to serialise
        returns: an ISO string if o is a datetime'''
        if isinstance(o, datetime):
            return o.isoformat()
        else:
            return super(MetadataEncoder, self).default(o)


class JSONLabNotebook(epyc.LabNotebook):
    '''A lab notebook that persists intself to a JSON file.'''

    def __init__( self, name, create = False, description = None ):
        '''Create a new JSON notebook, using the notebook's name
        as the JSON file. If this file exists, it will be opened
        and loaded unless create is True, in which case it will
        be erased.

        name: JSON file to persist the notebook to
        create: if True, erase existing file (defaults to False)
        description: free text description of the notebook'''
        super(epyc.JSONLabNotebook, self).__init__(name, description)

        # check for the file already existing
        if os.path.isfile(self.name()):
            # file exists, do we load it or create into it?
            if create:
                # empty the file
                with open(self.name(), 'w') as f:
                    f.write('')

                # preserve any description we were passed
                self._description = description
            else:
                # load notebook from file
                self._load(self.name())

    def isPersistent( self ):
        '''Return True to indicate the notebook is persisted to a JSON file.

        returns: True'''
        return True

    def commit( self ):
        '''Persist to disc.'''
        self._save(self.name())
        
    def _load( self, fn ):
        '''Retrieve the notebook from the given file.

        fn: the file name'''

        # if file is empty, create an empty notebook
        if os.path.getsize(fn) == 0:
            self._description = None
            self._results = dict()
            self._pending = dict()
        else:
            # load the JSON object
            with open(fn, "r") as f:
                s = f.read()

                # parse back into appropriate variables
                j = json.loads(s)
                self._description = j['description']
                self._pending = dict(j['pending'])
                self._results = j['results']

                # perform any post-load patching
                self.patch()

    def _patchDatetimeMetadata( self, res, mk ):
        '''Private method to atch an ISO datetime string to a datetime object
        for metadata key mk.

        res: results dict
        mk: metadata key'''
        t = res[epyc.Experiment.METADATA][mk]
        res[epyc.Experiment.METADATA][mk] = dateutil.parser.parse(t)
        
    def patch( self ):
        '''Patch the results dict. The default processes the Experiment.START_TIME
        and Experiment.END_TIME metadata fields back into Python datetime objects
        from ISO strings. This isn't strictly necessary, but it makes notebook
        data structure more Pythonic.'''

        for k in self._results.keys():
            ars = self._results[k]
            for res in ars:
                if isinstance(res, dict):
                    # not a pending result, patch its timing metadata
                    self._patchDatetimeMetadata(res, epyc.Experiment.START_TIME)
                    self._patchDatetimeMetadata(res, epyc.Experiment.END_TIME)
            
    def _save( self, fn ):
        '''Persist the notebook to the given file.

        fn: the file name'''

        # create JSON object
        j = json.dumps({ "description": self.description(),
                         "pending": self._pending.items(),
                         "results": self._results },
                       indent = 4,
                       cls = MetadataEncoder)
        
        # write to file
        with open(fn, 'w') as f:
            f.write(j)
            

    



# Simulation "lab" experiment management, parallel cluster version
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc

import numpy
import pickle
import dill

from ipyparallel import Client


class ClusterLab(epyc.Lab):
    '''A laboratory running over a cluster. Experiments are submitted to
    engines in the cluster for execution in parallel, with the experiments
    being performed asynchronously to allow for disconnection and subsequent
    retrieval of results. Combined with a persistent LabNotebook, this
    allows for fully decoupled access to an on-going computational
    experiment with piecewise retrieval.

    This class requires a cluster to already be set up and running, configured
    for persistent access, with access to the necessary code and libraries,
    and with appropriate security information available.'''

    def __init__( self, notebook = None, robust = True, url_file = None, profile = None, profile_dir = None, ipython_dir = None, context = None, debug = False, sshserver = None, sshkey = None, password = None, paramiko = None, timeout = 10, cluster_id = None, **extra_args ):
        '''Create an empty lab attached to the given cluster. most of the arguments
        are as expected by the ipyparallel.Client class, and are used to create the
        underlying connection to the cluster.

        Lab arguments:
           notebook: the notebook used to results (defaults to an empty LabNotebook)
           robust: if False, fails at individual job failures (defaults to True)

        Cluster client arguments:
           url_file: file containing connection information for accessing cluster
           profile: name of the IPython profile to use
           profile_dir: directory containing the profile's connection information
           ipython_dir: directory containing profile directories
           context: ZMQ context
           debug: whether to issue debugging information (defaults to False)
           sshserver: username and machine for ssh connections
           sshkey: file containing ssh key
           password: ssh password
           paramiko: True to use paramiko for ssh (defaults to False)
           timeout: timeout in seconds for ssh connection (defaults to 10s)
           cluster_id: string added to runtime files to prevent collisions'''
        super(epyc.ClusterLab, self).__init__(notebook)
        self._robust = robust
        
        # record all the connection arguments for later
        self._arguments = dict(url_file = url_file,
                               profile = profile,
                               profile_dir = profile_dir,
                               ipython_dir = ipython_dir,
                               context = context,
                               debug = debug,
                               sshserver = sshserver,
                               sshkey = sshkey,
                               password = password,
                               paramiko = paramiko,
                               timeout = timeout,
                               cluster_id = cluster_id,
                               **extra_args)
        self._client = None

        # connect to the cluster
        self.open()
        
        # make us use Dill as pickler by default
        self.use_dill()

    def open( self ):
        '''Connect to the cluster.'''
        if self._client is None:
            self._client = Client(**self._arguments)
        
    def close( self ):
        '''Close down the connection to the cluster.'''
        if self._client is not None:
            self._client.close()
            self._client = None
        
    def numberOfEngines( self ):
        '''Return the number of engines available to this lab.

        returns: the number of engines'''
        self.open()
        return len(self._client[:])

    def engines( self ):
        '''Return a list of the available engines.

        returns: a list of engines'''
        self.open()
        return self._client[:]

    def use_dill( self ):
        '''Make the cluster use Dill as pickler for transferring results.'''
        with self.sync_imports(quiet = True):
            import dill
        self._client.direct_view().use_dill()

    def sync_imports( self, quiet = False ):
        '''Return a context manager to control imports onto all the engines
        in the underlying cluster. This method is used within a with statement.

        quiet: if True, suppresses messages (defaults to False)
        returns: a context manager'''
        self.open()
        return self._client[:].sync_imports(quiet = quiet)
    
    def _mixup( self, ps ):
        '''Private method to mix up a list of values in-place using a Fisher-Yates
        shuffle (see https://en.wikipedia.org/wiki/Fisher-Yates_shuffle).

        ps: the array
        returns: the array, shuffled in-place'''
        for i in xrange(len(ps) - 1, 0, -1):
            j = int(numpy.random.random() * i)
            temp = ps[i]
            ps[i] = ps[j]
            ps[j] = temp
        return ps
     
    def runExperiment( self, e ):
        '''Run the experiment across the parameter space in parallel using
        all the engines in the cluster. This method returns immediately.
        The experiments are run asynchronously, with the points in the parameter
        space being explored randomly so that intermediate retrievals of results
        are more representative of the overall result. Put another way, for a lot
        of experiments the results available will converge towards a final
        answer, so we can plot them and see the answer emerge.        

        e: the experiment'''

        # create the parameter space
        space = self.parameterSpace()

        # only proceed if there's work to do
        if len(space) > 0:
            nb = self.notebook()
            
            # randomise the order of the parameter space so that we evaluate across
            # the space as we go along to make intermediate (incomplete) result
            # sets more representative of the overall result set
            ps = self._mixup(space)

            try:
                # connect to the cluster
                self.open()
            
                # submit an experiment at each point in the parameter space to the cluster
                view = self._client.load_balanced_view()
                jobs = view.map_async((lambda p: e.set(p).run()), ps)
                
                # record the mesage ids of all the jobs as submitted but not yet completed
                psjs = zip(ps, jobs.msg_ids)
                for (p, j) in psjs:
                    nb.addPendingResult(p, j)
            finally:
                # close our connection to the cluster and commit our pending results
                self.close()
                nb.commit()

    def _updateResults( self ):
        '''Update the jobs record with any newly-completed jobs.

        returns: the number of jobs completed at this call'''

        # sd: this may not be the most efficient way: may be better to
        # work out the ready jobs and then grab them all in one network transaction
        nb = self.notebook()
        n = 0
        try:
            self.open()
            for j in nb.pendingResults():
                try:
                    if self._client.get_result(j).ready():
                        # result is ready, grab it
                        r = (self._client.get_result(j).get())[0] # wrangle the list
                                                                  # that gets returned
                                                                  # into a single value
                        n = n + 1

                        # update the result in the notebook, cancelling
                        # the corresponding pending job
                        nb.addResult(r, j)
                except Exception as x:
                    if self._robust:
                        # we're running robustly, so elide the exception 
                        # (there seems to sometimes be a race condition
                        # the makes retrieval fail, but succeed later)
                        #print "Job id {id} inaccessible, will try again".format(id = j)
                        pass
                    else:
                        # we're not running robustly, pass on the exception
                        raise x
        finally:
            # whatever happens, commit changes to the notebook
            # and close the connection to the cluster 
            self.close()
            if n > 0:
                nb.commit()
        return n
                
    def results( self ):
        '''Return all the results we have at present.

        returns: a list of available results'''

        # grab any results that have come in recently
        self._updateResults()

        # collect the results into a list
        return self.notebook().results()

    def _availableResults( self ):
        '''Private method to return the number of results available.
        This does not update the results fetched from the cluster.

        returns: the number of available results'''
        return self.notebook().numberOfResults()

    def _availableResultsFraction( self ):
        '''Private method to return the fraction of results available, as a real number
        between 0 and 1. This does not update the results fetched from the cluster.

        returns: the fraction of available results'''
        tr = self.notebook().numberOfResults() + self.notebook().numberOfPendingResults()
        if tr == 0:
            return 0
        else:
            return (self._availableResults() + 0.0) / tr
    
    def readyFraction( self ):
        '''Test what fraction of results are available. This will change over
        time as the results come in.

        returns: the fraction from 0 to 1'''

        # grab any results that have come in recently
        self._updateResults()
        
        # return the fraction
        return self._availableResultsFraction()
    
    def ready( self ):
        '''Test whether all the results are available. This will change over
        time as the results come in.

        returns: True if all the results are available'''
        return (self.readyFraction() == 1)



# Initialisation for epyc package
#
# Copyright (C) 2016 Simon Dobson
#
# Licensed under the GNU General Public Licence v.2.0
#

from .experiment import Experiment
from .repeatedexperiment import RepeatedExperiment
from .summaryexperiment import SummaryExperiment

from .lab import Lab
from .clusterlab import ClusterLab

from .labnotebook import LabNotebook
from .jsonlabnotebook import JSONLabNotebook
from .sqlitelabnotebook import SqliteLabNotebook





# Simulation "lab notebook" for collecting results, in-memory version
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from epyc import *
from pandas import DataFrame


class LabNotebook(object):
    '''A "laboratory notebook" recording the results of a set of
    experiments conducted across a parameter space. The intention is
    to record all the metadata necessary to re-conduct the experiment.

    A notebook maps points nn a parameter space to a set of results. There
    may be multiple results mapped to the same point, to allow for
    repetition of experiments: the notbook can re-cquire all, or only the
    most recent, results for a given parameter point.

    Notebooks are immutable: once entered, a result can't be deleted
    or changed.

    Notebooks support "pending" results, llowing us to record experiments
    in progress. A pending result can be finalised by providing it with a
    value, or can be deleted.

    Notebooks support both len() and iterator access, both referring to their
    complete results.'''

    def __init__( self, name = None, description = None ):
        '''Create an empty notebook.

        name: the notebook's name
        description a free text description'''
        self._name = name
        self._description = description
        self._results = dict()
        self._pending = dict()

    def name( self ):
        '''Return the name of the notebook. If the notebook is persistent,
        this likely relates to its storage in some way (for example a
        file name).

        returns: the notebook name'''
        return self._name

    def description( self ):
        '''Return the free text description of the notebook.

        returns: the notebook description'''
        return self._description

    def isPersistent( self ):
        '''By default notebooks are not persistent.

        returns: False'''
        return False

    def commit( self ):
        '''Commit to persistent form. By default does nothing. This should
        be called periodically to save intermediate results: it may happen
        automatically in some sub-classes, depending on their implementation.'''
        pass
    
    def _parametersAsIndex( self, ps ):
        '''Private method to turn a parameter dict into a string suitable for
        keying a dict.

        ps: the parameters as a hash
        returns: a string key'''
        k = ""
        for p in sorted(ps.keys()):       # normalise the parameters
            v = ps[p]
            k = k + "{p}=[[{v}]];".format(p = p, v = v)
        return k
        
    def addResult( self, results, jobids = [] ):
        '''Add a result. This should be a dict as returned from an instance of Experiment,
        that contains metadata, parameters, and result. Results cannot be overridden, as
        notebooks are immutable: adding more results simply adds another result set.

        If results is a list of result dicts, they are added in order: the last result in
        the list is assumed to be the latest.

        If the jobid is present, this result resolves the corresponding pending result.
        As with result, jobid can be a list. (The two lists need not be the same
        length, to allow for experiments that return lists of result dicts.)

        result: the results (or a list of results)
        jobid: the pending result job id(s) (defaults to no jobs)'''

        # make single parameters into lists where appropriate 
        if not isinstance(results, list):
            results = [ results ]
        if not isinstance(jobids, list):
            jobids = [ jobids ]

        # add each result
        for result in results:
            k = self._parametersAsIndex(result[Experiment.PARAMETERS])

            # retrieve or create the result list
            if k in self._results.keys():
                rs = self._results[k]
            else:
                rs = []
                self._results[k] = rs

            # store the result
            rs.insert(0, result)

        # if there is are job ids provided, cancel the corresponding pending jobs
        if jobids is not []:
            for jobid in jobids:
                if (jobid in rs) and (jobid in self._pending.keys()):
                    # delete job id from current results
                    j = rs.index(jobid)
                    del rs[j]
                    
                    # ...and from the set of pending results
                    del self._pending[jobid]
                else:
                    # we've screwed-up the internal data structures
                    raise RuntimeError('Internal structure error for {j} -> {ps}'.format(j = jobid,
                                                                                         ps = k))
        
    def addPendingResult( self, ps, jobid ):
        '''Add a "pending" result that we expect to get results for.

        ps: the parameters for the result
        jobid: an identifier for the pending result'''
        k = self._parametersAsIndex(ps)

        # retrieve or create the result list
        if k in self._results.keys():
            rs = self._results[k]
        else:
            rs = []
            self._results[k] = rs

        # append the pending result's jobid
        rs.insert(0, jobid)

        # map job id to parameters to which it refers
        self._pending[jobid] = k

    def cancelPendingResult( self, jobid ):
        '''Cancel a particular pending result. Note that this only affects the
        notebook's record, not any job running in a lab.

        jobid: job id for pending result'''
        if jobid in self._pending.keys():
            k = self._pending[jobid]
            del self._pending[jobid]
            if k in self._results.keys():
                rs = self._results[k]
                j = rs.index(jobid)
                del rs[j]
            else:
                # we've screwed-up the internal data structures
                raise RuntimeError('Internal structure error for {j} -> {ps}'.format(j = jobid,
                                                                                     ps = k))
        else:
            # no such job
            # sd: should this just fail silently?
            raise KeyError('No pending result with id {j}'.format(j = jobid))
        
    def pendingResultsFor( self, ps ):
        '''Retrieve a list of all pending results associated with the given parameters.

        ps: the parameters
        returns: a list of pending result job ids, which may be empty'''
        k = self._parametersAsIndex(ps)
        if k in self._results.keys():
            # filter out pending job ids, which can be anything except dicts
            return [ j for j in self._results[k] if not isinstance(j, dict) ]
        else:
            return []
                
    def cancelAllPendingResultsFor( self, ps ):
        '''Cancel all pending results for the given parameters. Note that
        this only affects the notebook's record, not any job running in a lab.

        ps: the parameters'''
        k = self._parametersAsIndex(ps)

        if k in self._results.keys():
            # remove from results
            rs = self._results[k]
            js = [ j for j in rs if not isinstance(j, dict) ]
            self._results[k] = [ rc for rc in rs if isinstance(rc, dict) ]

            # ...and from pending jobs list
            for j in js:
                del self._pending[j]

    def cancelAllPendingResults( self ):
        '''Cancel all pending results. Note that this only affects the
        notebook's record, not any job running in a lab.'''
        for k in self._results.keys():
            rs = self._results[k]
            self._results[k] = [ j for j in rs if isinstance(j, dict) ]
        self._pending = dict()

    def pendingResults( self ):
        '''Return the job ids of all pending results.

        returns: a list of job ids, which may be empty'''
        return self._pending.keys()
        
    def resultsFor( self, ps ):
        '''Retrieve a list of all results associated with the given parameters.

        ps: the parameters
        returns: a list of results, which may be empty'''
        k = self._parametersAsIndex(ps)
        if k in self._results.keys():
            # filter out pending job ids, which can be anything except dicts
            return [ res for res in self._results[k] if isinstance(res, dict) ]
        else:
            return []

    def latestResultsFor( self, ps ):
        '''Retrieve only the latest result for the given parameters.

        ps: the parameters
        returns: a single result, or None if there are none'''
        rs = self.resultsFor(ps)
        if rs is None:
            return None
        else:
            if len(rs) == 0:
                return None
            else:
                return rs[0]
        
    def results( self ):
        '''Return a list of all the results currently available. This
        excludes pending results. Results are returned as a single flat
        list, so any repetition structure is lost.

        returns: a list of results'''
        rs = []
        for k in self._results.keys():
            # filter out pending job ids, which can be anything except dicts
            ars = [ res for res in self._results[k] if isinstance(res, dict) ]
            rs.extend(ars)
        return rs

    def numberOfResults( self ):
        '''Return the number of results we have.

        returns: number of results available'''
        return len(self.results())

    def numberOfPendingResults( self ):
        '''Return ther number of addiitonal results we expect.

        returns: number of pending results'''
        return len(self._pending.keys())
    
    def __len__( self ):
        '''The length of a notebook is the number of results it currently
        has available. A synonym for numberOfResults().

        returns: the number of results available'''
        return self.numberOfResults()
    
    def __iter__( self ):
        '''Return an iterator over the results available.

        returns: an iteration over the results'''
        return self.results().__iter__()
    
    def dataframe( self ):
        '''Return the results as a pandas DataFrame. Note that there is a danger
        of duplicate labels here, for example if the results contain a value
        with the same name as one of the parameters. To resolve this, parameter names
        take precedence over metadata values, and result names take precedence over
        parameter names.

        returns: the parameters, results, and metadata in a DataFrame'''

        def extract( r ):
            rs = r[Experiment.METADATA].copy()
            rs.update(r[Experiment.PARAMETERS])
            rs.update(r[Experiment.RESULTS])
            return rs

        return DataFrame.from_dict(map(extract, self.results()))
    



# Test suite
#
# Copyright (C) 2016 Simon Dobson
#
# Licensed under the GNU General Public Licence v.2.0
#

import unittest
from .experiments import *
from .repeatedexperiments import *
from .summaryexperiments import *
from .labs import *
from .clusterlabs import *
from .notebooks import *
from .jsonnotebooks import *
from .sqlitenotebooks import *

experimentsSuite = unittest.TestLoader().loadTestsFromTestCase(ExperimentTests)
repeatedExperimentsSuite = unittest.TestLoader().loadTestsFromTestCase(RepeatedExperimentTests)
summaryExperimentsSuite = unittest.TestLoader().loadTestsFromTestCase(SummaryExperimentTests)
notebooksSuite = unittest.TestLoader().loadTestsFromTestCase(LabNotebookTests)
jsonnotebooksSuite = unittest.TestLoader().loadTestsFromTestCase(JSONLabNotebookTests)
labsSuite = unittest.TestLoader().loadTestsFromTestCase(LabTests)
clusterlabsSuite = unittest.TestLoader().loadTestsFromTestCase(ClusterLabTests)

suite = unittest.TestSuite([ experimentsSuite,
                             repeatedExperimentsSuite, summaryExperimentsSuite,
                             notebooksSuite, jsonnotebooksSuite,
                             labsSuite, clusterlabsSuite ])

if __name__ == '__main__':
    unittest.TextTestRunner(verbosity = 2).run(suite)



# Tests of repeated experiments combinator
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from epyc import *

import unittest
import numpy
import numpy.random


class SampleExperiment1(Experiment):
    
    def do( self, params ):
        return dict(result = params['x'])

    
class RepeatedExperimentTests(unittest.TestCase):

    def setUp( self ):
        self._lab = Lab()
        
    def testRepetitionsOnePoint( self ):
        '''Test we can repeat an experiment at a single point'''
        N = 10
        
        self._lab['x'] = [ 5 ]
        
        e = SampleExperiment1()
        er = RepeatedExperiment(e, N)

        self._lab.runExperiment(er)
        self.assertTrue(er.success())
        
        results = self._lab.results()
        self.assertEqual(len(results), N * len(self._lab['x']))
        for res in self._lab.notebook().resultsFor(dict(x = self._lab['x'][0])):
            self.assertEqual(res[Experiment.RESULTS]['result'], self._lab['x'][0])
            
    def testRepetitionsMultiplePoint( self ):
        '''Test we can repeat an experiment across a parameter space'''
        N = 10
        
        self._lab['x'] = [ 5, 10, 15 ]
        
        e = SampleExperiment1()
        er = RepeatedExperiment(e, N)

        self._lab.runExperiment(er)
        results = self._lab.results()
        self.assertEqual(len(results), N * len(self._lab['x']))

        for x in range(len(self._lab['x'])):
            for res in self._lab.notebook().resultsFor(dict(x = x)):
                self.assertIn(res[Experiment.PARAMETERS]['x'], x)
                self.assertEqual(res[Experiment.RESULTS]['result'], x)





# Tests of Sqlite notebooks
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

from epyc import *

import unittest
import os
from tempfile import NamedTemporaryFile


class SqlliteLabNotebookTests(unittest.TestCase):

    def testCreateDatabaseInMemory( self ):
        '''Test creation of notebook using SQLite database in memory'''
        db = SqliteLabNotebook('test')
        self.assertFalse(db.isPersistent())
        self.assertEqual(db.name(), 'test')

    def testCreateDatabase( self ):
        '''Test creation of notebook using SQLite database in a file'''
        tf = NamedTemporaryFile()
        tf.close()
        fn = tf.name
        
        db = SqliteLabNotebook('test', fn)
        self.assertTrue(db.isPersistent())
        self.assertEqual(db.name(), 'test')
        self.assertTrue(os.path.isfile(fn))
        os.remove(fn)



# Experiment combinator to run the same experiment several times and
# return only the summary statistics
#
# Copyright (C) 2016 Simon Dobson
# 
# Licensed under the GNU General Public Licence v.2.0
#

import epyc


class RepeatedExperiment(epyc.Experiment):
    '''A experiment combinator that takes a "base" experiment and runs it
    several times. This means you can define a single experiment separate
    from its repeating logic.

    When run, a repeated experiment runs a number of repetitions of the underlying
    experiment at the same point in the parameter space. The result of the
    repeated experiment is the list of results from the underlying experiment.
    If the underlying experiment returns a list of results, these are all
    flattened into a single list.'''

    def __init__( self, ex, N ):
        '''Create a repeated version of the given experiment.

        ex: the underlying experiment
        N: the number of repetitions to perform'''
        super(epyc.RepeatedExperiment, self).__init__()
        self._experiment = ex
        self._N = N

    def experiment( self ):
        '''Return the underlying experiment.

        returns: the underlying experiment'''
        return self._experiment

    def repetitions( self ):
        '''Return the number of repetitions of the underlying experiment
        we expect to perform.

        returns: the number of repetitions'''
        return self._N

    def do( self, params ):
        '''Perform the number of repetitions we want.

        params: the parameters to the experiment
        returns: a list of result dicts'''
        self.experiment().set(params)
        N = self.repetitions()
        results = []
        for i in xrange(N):
            res = self.experiment().run()
            if isinstance(res, list):
                results.extend(res)
            else:
                results.append(res)
        return results

    def report( self, params, meta, res ):
        '''Return just the results, don't add any more metadata.

        params: the parameters we ran under
        meta: the metadata for this run
        res: the list of results
        returns: the list of results'''
        return res
    
